#include <qapplication.h>
#include <qmainwindow.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <qfiledialog.h>
#include <qtimer.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdragobject.h>
#include <qfile.h>

#include <stdlib.h>

#include "QtScr.h"

QtScr::QtScr()
{
	// Contruccin de las tablas de Paper/Ink
	for( int idx = 0; idx < 256; idx++ )
	{
		Ink[idx]   = idx & 0x07;
		Paper[idx] = (idx >> 3) & 0x07;
		if( idx & 0x40 )  // Hay Bright???
		{
			Ink[idx]   |= 0x08;
			Paper[idx] |= 0x08;
		}
		Ink[idx] += 32;
		Paper[idx] += 32;
	}

	// Construccin de la tabla de direcciones de la pantalla
	for( int linea = 0; linea < 24; linea++ )
	{
		int idx, lsb, msb, addr;
		lsb = ((linea & 0x07) << 5);
		msb = linea & 0x18;
		addr = msb * 256 + lsb;
		idx = linea * 8;
		for( int scan = 0; scan < 8; scan++, addr += 256 )
			ScrAddr[scan + idx] = addr;
	}
	   
	pScr = new unsigned char[6144];
	pAttr = new unsigned char[768];
	for( int idx = 0; idx < 768; idx++ )
		pAttr[idx] = 0;

	flashOn = 0; // inicializamos el ciclo de flash
	border = true;  // color blanco por defecto para el borde

	pm = new QPixmap;
    doble = new QPixmap;
    matriz.scale((double)2.0, (double)2.0);

	img = new QImage(256, 192, 8, 64, QImage::LittleEndian);
	img->setColor(32, qRgb(0, 0, 0));        /* Black */
	img->setColor(33, qRgb(0, 0, 192));      /* Blue */
	img->setColor(34, qRgb(192, 0, 0));      /* Red */
	img->setColor(35, qRgb(192, 0, 192));    /* Magenta */
	img->setColor(36, qRgb(0, 192, 0));      /* Green */
	img->setColor(37, qRgb(0, 192, 192));    /* Cyan */
	img->setColor(38, qRgb(192, 192, 0));    /* Yellow */
	img->setColor(39, qRgb(192, 192, 192));  /* White */
	img->setColor(40, qRgb(0, 0, 0));        /* Black */
	img->setColor(41, qRgb(0, 0, 255));      /* Bright Blue */
	img->setColor(42, qRgb(255, 0, 0));      /* Bright Red */
	img->setColor(43, qRgb(255, 0, 255));    /* Bright Magenta */
	img->setColor(44, qRgb(0, 255, 0));      /* Bright Green */
	img->setColor(45, qRgb(0, 255, 255));    /* Bright Cyan */
	img->setColor(46, qRgb(255, 255, 0));    /* Bright Yellow */
	img->setColor(47, qRgb(255, 255, 255));  /* Bright White */

    timer = new QTimer(this);
    connect( timer, SIGNAL(timeout()), SLOT(refreshScreen()) );

	// Crear el men Popup
	menu = new QPopupMenu(this, "QtScr Menu");
	CHECK_PTR( menu );
	menu->insertItem("Cargar imagen...", this, SLOT(CargaScr()));
	menu->insertItem("Cambiar color del borde", this, SLOT(cambioBorder()));
	attrMenu = menu->insertItem("Pantalla sin atributos", this,
								SLOT(cambiaAtributos()), Key_Space);
	dobleMenu = menu->insertItem("&Doble tamao", this, SLOT(DobleTamano()),
								 Key_D);
	menu->insertSeparator();
	menu->insertItem("Acerca de...", this, SLOT(AcercaDe()));
	menu->insertSeparator();
	menu->insertItem("Salir", this, SLOT(quit()));
	menu->setCheckable( TRUE );

	noVerAttr = false;   // por defecto se ven los atributos

	// Redimensionar la ventana al tamao del Spectrum (256x192)
	// ms un borde por cada lado de 32 pixels de ancho
    resize( 320, 256 );

	// Inicializar el color del fondo
	setBackgroundColor( Qt::lightGray );

	// Activar funcionalidad de Drag & Drop
    setAcceptDrops(true);

    dobleTam = false;
}

QtScr::~QtScr()
{
	delete []pScr;
	delete []pAttr;
	delete pm;
	delete img;
}

/*
 * Cambiar el cliclo de flash cada 16/50 de segundo (aka 320 ms)
 */
void QtScr::refreshScreen(void)
{
  flashOn = !flashOn;
  repaint( FALSE );
}

/*
 * Gestiona las pulsaciones de los botones del ratn
 */
void QtScr::mousePressEvent( QMouseEvent *event )
{
	// Si se ha pulsado el botn central del ratn, mostrar el men
	if( event->button() == Qt::RightButton )
		menu->popup(event->globalPos());
}

/*
 * Rutina de dibujo de la pantalla
 */
void QtScr::paintEvent( QPaintEvent * )
{
	unsigned int paper, ink;
	int addr;
	unsigned char val, flash, attr, *pixel;
	register unsigned char bit;
	register unsigned char mask;

	for( int cordy = 0, nAttr = 0; cordy < 192; cordy++ )
	{
		addr = ScrAddr[cordy];
        nAttr = (cordy & 0xF8) << 2;
		pixel = img->scanLine(cordy);
		for( int columna = 0; columna < 32; columna++ )
		{
			if( noVerAttr )  // pintar la pantalla sin atributos
			{
				ink = 32;
				paper = 39;
				flash = 0;
			}
			else             // pintar la pantalla en colorines
			{
				attr = pAttr[nAttr++];
				ink = Ink[attr];
				paper = Paper[attr];
				flash = (attr & 0x80) && flashOn;
			}
			val = pScr[addr++];

			for( mask = 0x80; mask; mask >>= 1, pixel++ )
			{
				bit = ((val & mask) ? 1 : 0) ^ flash;
				if( bit )
					*pixel = ink;
				else
					*pixel = paper;
			}
		}
	}

	pm->convertFromImage(*img);
    if( dobleTam )
    {
        *doble = pm->xForm( matriz );
        bitBlt( this, 64, 64, doble );
    }
    else
        bitBlt( this, 32, 32, pm );
}

/*
 * Se ocupa de la funcionalidad del Drag & Drop
 */
void QtScr::dropEvent( QDropEvent *dndzone )
{
	QStrList list;
	QString last;

    QTextDrag::decode(dndzone, last);

	if ( last.findRev(".scr", -1, FALSE) != -1 )
	{
		timer->stop();
		last.remove(0, 5);
		leerFichero(last.data());
		if( noVerAttr )
			cambiaAtributos();
		else
		{
			repaint( FALSE );
			timer->start( 320 );
			int pos = last.findRev('/', -1 , FALSE);
			QString titulo = "QtScr: ";
			titulo += last.right( last.length() - ++pos );
			setCaption(titulo.data());
		}

	}
	else
		QMessageBox::information(0, "Error!", "El fichero no es un volcado\n"
                                 "de la pantalla del Spectrum");
}

void QtScr::dragEnterEvent( QDragEnterEvent *qdragenterevent )
{
    qdragenterevent->accept( QTextDrag::canDecode(qdragenterevent) );
}

bool QtScr::leerFichero(const char *filename)
{
	QFile fp(filename);
	int nRead;

	if( (fp.open(IO_ReadOnly)) == false )
	{
		QMessageBox::warning(0, "Error!", "No se puede abrir el fichero");
		return false;
	}

	nRead = fp.readBlock((char *)pScr, 6144);
//	if( nRead != 6144 )
//	{
//		printf("Warning: Se han ledo %d datos de pantalla\n", nRead);
//		printf("         desde el fichero \"%s\"\n", filename);
//	}

	nRead += fp.readBlock((char *)pAttr, 768);
	if( nRead != 6912 )
	{
//		printf("Warning: Se han ledo %d datos de atributos\n", nRead);
//		printf("         desde el fichero \"%s\"\n", filename);
        QMessageBox::warning(0, "Atencin!", 
                             "La pantalla se ha ledo con errores");
	}

	fp.close();

	return true;
}

void QtScr::CargaScr(void)
{
	QString filename;

	timer->stop();
	filename = QFileDialog::getOpenFileName(0, "*.scr", this,
											"Abrir fichero de pantalla");
	if( !filename.isNull() )
	{
		leerFichero(filename.data());
		int pos = filename.findRev('/', -1 , FALSE);
		QString titulo = "QtScr: ";

		if( pos != -1 )
			titulo += filename.right( filename.length() - ++pos );
		else
			titulo += filename.data();

		setCaption(titulo.data());
	}
	else
		return;

	if( noVerAttr )
		cambiaAtributos();
	else
	{
		repaint( FALSE );
		timer->start( 320 );
	}
}

void QtScr::cambioBorder(void)
{
	timer->stop();

	// Cambiar el color del fondo
    border = !border;
    if( border )
        setBackgroundColor( Qt::lightGray );
    else
        setBackgroundColor( Qt::black );

	repaint( TRUE );

	timer->start( 320 );
}

void QtScr::AcercaDe(void)
{
		QMessageBox::about(this, "Acerca de QtScr",
						   "          QtScr v0.8\n"
                           "Visualizador de pantallas del Spectrum\n"
                           "(c) 1999-2003 Jos Luis Snchez Villanueva\n"
                           "jsanchez@todounix.homeip.net\n");
}

void QtScr::cambiaAtributos(void)
{
	if( noVerAttr )
	{
		noVerAttr = false;
		menu->setItemChecked(attrMenu, FALSE);
		timer->start(320);
	}
	else
	{
		noVerAttr = true;
		menu->setItemChecked(attrMenu, TRUE);
		timer->stop();
	}
	repaint(FALSE);
}

void QtScr::DobleTamano(void)
{
    if( dobleTam )
    {
        dobleTam = false;
        menu->setItemChecked(dobleMenu, FALSE);
        resize(320, 256);
    }
    else
    {
        dobleTam = true;
        menu->setItemChecked(dobleMenu, TRUE);
        resize(640, 512); 
    }
}

void QtScr::quit(void)
{
	exit(0);
}
